#pragma once

#include <cstddef>

// A struct for representing a string. Unlike a C string
// ('\0'-terminated char array), this struct records the length of the
// string in the `size_` field, so `data_` can contain '\0's.
//
// (We will also '\0'-terminate it so that it can be passed to C
// functions that expect that, but C++ doesn't need that.)
//
// For this struct to be valid, some conditions (invariants) need to
// hold:
//
//  1. `capacity_` is 0 if and only if `data_` is null.
//
//  2. If `capacity_` is non-zero then it contains the actual allocated
//  size of the array object pointed to by `data_.`
//
//  3. `size_ + 1 <= capacity_`
//
//  4. The first `size_` elements of `data_` are initialized, and
//  `data_[size_] == '\0'`.
//
// The underscores on the member variable names (“member variable” is
// C++-speak for “field”) are a convention meaning that they are
// *private*, which means that client code of this header should never
// access them directly, but always via the functions below. This is to
// ensure that the invariants above are never broken. C++ understand
// privacy and can prevent the client from accessing members, but we
// aren't going to use that feature yet.
struct Owned_string
{
    /*
     * Constructors.
     *
     * These are used to initialize a new `Owned_string` object, possibly
     * from some arguments. Every `Owned_string` object that is constructed
     * will be destroyed automatically using the destructor `~Owned_string`
     * declared below.
     */

    // Initializes `this` to the empty string.
    Owned_string();

    // Initializes `this` to a copy of the given C ('\0'-terminated)
    // string.
    Owned_string(const char*);

    // Initializes `this` to a copy of the array starting at `begin` and
    // ending at (not including) `end`. This array may contain '\0's, and
    // they will be copied too.
    Owned_string(const char* begin, const char* end);

    // Copy constructor: initializes `this` to be a copy of the
    // argument.
    Owned_string(Owned_string const&);

    // Move constructor: initializes `this` by stealing the argument's
    // memory, leaving it valid but empty.
    Owned_string(Owned_string&&) noexcept;

    /*
     * Destructor.
     */

    // C++ calls this automatically whenever a `Owned_string` object needs to
    // be destroyed (such as when it goes out of scope).
    ~Owned_string();

    /*
     * Assignment “operators”.
     */

    // Assigns the contents of the argument to `this`. This may reuse
    // `this`'s memory or reallocate.
    Owned_string& operator=(Owned_string const&);

    // Steals the argument's memory, assigning it to `this` and leaving
    // the argument object valid but empty.
    Owned_string& operator=(Owned_string&&) noexcept;

    /*
     * Non-lifecycle `Owned_string` operations.
     */

    // Returns whether `this` is the empty string.
    bool empty() const;

    // Returns the number of characters in `this`. This does not
    // depend on '\0' termination, and internal '\0's are allowed.
    size_t size() const;

    // Returns the character of `this` at the given index.
    //
    // ERROR: If `index >= this->size()` then the behavior is
    // undefined.
    char operator[](size_t index) const;

    // Returns a reference to the character of `this` at the given index.
    //
    // ERROR: If `index >= this->size()` then the behavior is
    // undefined.
    char& operator[](size_t index);

    // Adds character `c` to the end of `this`. This may cause pointers
    // returned by previous calls to `this->operator[]` or `this->c_str`
    // to become invalidated.
    void push_back(char c);

    // Removes the last character of the string.
    //
    // ERROR: UB if `Owned_string(this)`.
    void pop_back();

    // Expands the capacity, if necessary, to hold `additional_cap` more
    // characters.
    void reserve(size_t additional_cap);

    // Swaps the contents of two strings without copying the actual
    // characters.
    void swap(Owned_string&);

    // Appends another string onto this string.
    Owned_string& operator+=(Owned_string const& that);

    // Appends a C string onto this string.
    Owned_string& operator+=(const char* that);

    // Returns a pointer to the content of this string as a C-style string.
    // Note that internal '\0's will make `strlen(String_c_str(s))` less than
    // `String_size(s)`, which doesn't depend on the content of the string.
    const char* c_str() const;

    /*
     * Some operations that were hard to define in lec09-String.h-style.
     */

    // Returns a pointer to the first character of the string.
    const char* begin() const;
    char* begin();

    // Returns a pointer to the first character past the end of the
    // string.
    const char* end() const;
    char* end();

    /*
     * PRIVATE HELPER FUNCTIONS.
     *
     * These are functions that are useful for implementing the
     * functions above. They should not be called by clients.
     * Thus, they are documented in the implementation file, not
     * in this header.
     */

private:
    void set_empty_();
    void ensure_capacity_(size_t min_cap);
    void append_range_(const char* begin, const char* end);

    /*
     * Data members -- clients can't touch these.
     *
     * These are where the data is actually stored---everything above
     * here declares operations (member functions), and below is data.
     */

    size_t size_;
    size_t capacity_;
    char*  data_;
};

/*
 * These are free functions. Some of them *could* be defined as members,
 * but it improves encapsulation to minimize the number of members of
 * possible.
 */

// Appends two strings, returning a new string.
Owned_string operator+(Owned_string const&, Owned_string const&);

// Appends two strings, returning a new string.
Owned_string operator+(Owned_string const&, const char*);

// Appends two strings, returning a new string.
Owned_string operator+(const char*, Owned_string const&);

// Steals the left string's memory to append the right and return the
// result.
Owned_string operator+(Owned_string&&, Owned_string const&);

// Steals the left string's memory to append the right and return the
// result.
Owned_string operator+(Owned_string&&, const char*);

